#include <mpi.h>
#include <vector>
#include "defs.h"
#include "alloc.h"
#include "timer.h"

// For intra-node only cases, split the World communicator to node-local ones
#define WITH_PER_NODE_COMM

void peer2peer_exchange(const std::vector<int> &params, std::vector<std::vector<double>> &output) {
    int rank = params[0];
    int size = params[1];
    int distance = params[2];
    int npeers_per_node = params[3];
    bool use_gpu_memory = (bool)params[4];
    bool is_intra_node = (2 * distance <= npeers_per_node) && (npeers_per_node % distance == 0);
    std::string kind = (is_intra_node ? "intra-node" : "cross-node");
    int r = rank % (2 * distance);
    bool do_recv_first = false;
    int peer = rank + distance;
    if (r >= distance) {
        do_recv_first = true;
        peer = rank - distance; 
    } 
    MPI_Comm comm;
	if (is_intra_node) {
#ifdef WITH_PER_NODE_COMM
	    MPI_Comm_split_type(MPI_COMM_WORLD, MPI_COMM_TYPE_SHARED, 0, MPI_INFO_NULL, &comm);
        peer = peer % npeers_per_node;
#else
		MPI_Comm_dup(MPI_COMM_WORLD, &comm);
#endif
	} else {
		MPI_Comm_dup(MPI_COMM_WORLD, &comm);
	}
    if (peer >= size || peer < 0) {
	    MPI_Comm_free(&comm);
        return;  
    }
    int niters = 10;
    int nwarmup = 3;
    size_t bufsize = (1<<26) * sizeof(real);
    size_t cachesize = 1<<23;
    char *buf = nullptr;
    if (use_gpu_memory) {
        buf = sys::gpu_alloc(bufsize + cachesize);
    } else {
        buf = sys::cpu_alloc(bufsize + cachesize);
    }
    std::vector<double> result;
    {
        timer t(rank2str(rank) + "Peer-to-peer bandwidth:", bufsize * 2, niters, result);
        t.comment << kind << " with distance=" << distance;
        if (!is_intra_node || use_gpu_memory) { // FIXME it is a bug workaround -- in CPU mode intra-node exch may deadlock
            for (int i = 0; i < niters + nwarmup; i++) {
                if (i == nwarmup)
                    t.start();
                if (do_recv_first) {
                    //printf("rank=%d: MPI_Recv(buf + (i%%2) * cachesize, bufsize, MPI_BYTE, peer=%d, 0, comm, MPI_STATUS_IGNORE);\n", rank, peer);
                    MPI_Recv(buf + (i%2) * cachesize, bufsize, MPI_BYTE, peer, 0, comm, MPI_STATUS_IGNORE);
                    //printf("rank=%d: MPI_Send(buf + (i%%2) * cachesize, bufsize, MPI_BYTE, peer=%d, 0, comm);\n", rank, peer);
                    MPI_Send(buf + (i%2) * cachesize, bufsize, MPI_BYTE, peer, 0, comm);
                } else {
                    //printf("rank=%d: MPI_Send(buf + (i%%2) * cachesize, bufsize, MPI_BYTE, peer=%d, 0, comm);\n", rank, peer);
                    MPI_Send(buf + (i%2) * cachesize, bufsize, MPI_BYTE, peer, 0, comm);
                    //printf("rank=%d: MPI_Recv(buf + (i%%2) * cachesize, bufsize, MPI_BYTE, peer=%d, 0, comm, MPI_STATUS_IGNORE);\n", rank, peer);
                    MPI_Recv(buf + (i%2) * cachesize, bufsize, MPI_BYTE, peer, 0, comm, MPI_STATUS_IGNORE);
                }
            }
        } else {
            t.comment << " SKIPPED!";
            t.erase();
        }
    }     
    if (use_gpu_memory) {
        sys::gpu_free(buf);
    } else {
        sys::cpu_free(buf);
    }
	MPI_Comm_free(&comm);
    output.push_back(result);
}

